// Main Container Function, that calls itself
(function () {
	if (typeof mraid !== "undefined") {
		console.log("mraid.js MRAID is already loaded, aborting");
		return;
	}

	//////////////////////////////////////////////////////////////////////////////////////
	//                                  MRAID VERSION                                   //
	var mraidVersion = "3.0";
	//                                                                                  //
	//////////////////////////////////////////////////////////////////////////////////////
    var apiVersion = '2.0';
	//////////////////////////////////////////////////////////////////////////////////////
	//                                    EVENTS                                        //
	/** MRAID 3.0 Events*/
	var EVENT_ERROR = "error";
	var EVENT_READY = "ready";
	var EVENT_SIZECHANGE = "sizeChange";
	var EVENT_STATECHANGE = "stateChange";
	var EVENT_EXPOSURECHANGE = "exposureChange";
	var EVENT_AUDIOVOLUMECHANGE = "audioVolumeChange";
	var EVENT_VIEWABLECHANGE = "viewableChange";   /** MRAID 3.0 deprecated */


	/** non-MRAID 3.0 Events*/
	var EVENT_SUCCESS = "success";
	var EVENT_WARNING = "warning";
	var EVENT_VISXONSCROLL = "visxOnScroll";
	var EVENT_VISXONBACKPRESS = "visxOnBackPressed";
	var EVENT_VISX_STICKY_IS_CLOSED= "visxStickyIsClosed";
	var EVENT_VISX_STICKY_IS_DISPLAYED = "visxStickyIsDisplayed";
	var EVENT_VISX_TO_STICKY = "visxToSticky";
	var EVENT_VISX_TO_STICKY_CLOSE = "visxToStickyClose";
	var EVENT_VISX_TO_STICKY_CLOSE_BUTTON = "visxToStickyCloseButton";
	//                                                                                  //
    //////////////////////////////////////////////////////////////////////////////////////

	//////////////////////////////////////////////////////////////////////////////////////
	//                                  MRAID STATES                                    //
	var STATE_LOADING = "loading";
	var STATE_HIDDEN = "hidden";
	var STATE_DEFAULT = "default";
	var STATE_EXPANDED = "expanded";
	var STATE_RESIZED = "resized";
	var states = [STATE_LOADING, STATE_HIDDEN, STATE_DEFAULT, STATE_EXPANDED, STATE_RESIZED];
	var state = STATE_LOADING;
	//                                                                                  //
	//////////////////////////////////////////////////////////////////////////////////////

	//////////////////////////////////////////////////////////////////////////////////////
	//                                 MRAID PROPERTIES                                 //
	var orientationProperties = {
		allowOrientationChange: true,
		forceOrientation: "none"
	};
	var expandProperties = {}, listeners, placementType, state, states, viewable, __slice = Array.prototype.slice, supportedFeatures = {};

	// Properties which define the behavior of a resizeable ad.
	var resizeProperties = null;
	var resizePropertyValidators = {
		width: function (v) {
			return !isNaN(v) && v >= 0;
		},
		height: function (v) {
			return !isNaN(v) && v >= 0;
		},
		offsetX: function (v) {
			return !isNaN(v);
		},
		offsetY: function (v) {
			return !isNaN(v);
		},
		/** any value is okay, empty and unknown ones will fall back to a default*/
		customClosePosition: function (v) { return true; },
		allowOffscreen: function (v) {
			return (typeof v === 'boolean');
		}
	};

	var screenSize = {
		width: 0,
		height: 0
	};

	var defaultPosition = {
		x: 0,
		y: 0,
		width: 0,
		height: 0
	};
	var viewable = false;

	var FORCEORIENTATION_PORTRAIT = "portrait";
	var FORCEORIENTATION_LANDSCAPE = "landscape";
	var FORCEORIENTATION_NONE = "none";

	var FEATURES = ['sms', 'tel', 'calendar', 'storePicture', 'inlineVideo'];

	var vpaidObject;
	//                                                                                  //
	//////////////////////////////////////////////////////////////////////////////////////

	//////////////////////////////////////////////////////////////////////////////////////
	//                               NON-MRAID PROPERTIES                               //
	var placementEffect;
	var placementType = "inline";
    	var isUniversalAd = false;
	var isExposureChangeEnabled = true;
	var isMediationAdView = false;
	var canUseOnScrollEvent = false;
	var webViewDimension = {
		width: 0,
		height: 0
	};
	var viewportDimension = {
		width: 0,
		height: 0
	};
	var absSize = {
		x: 0,
		y: 0,
		width: 0,
		height: 0
	};
	var currentAppOrientationProperties = {
        locked: true,
        orientation: "none"
    };
	var density = 1;
	var volume = 0;
	listeners = {};

	// Properties which define the behavior of a sticky ad.
    var stickyProperties = null;
    var stickyPropertyValidators = {
    	threshold: function (v) {
    		return !isNaN(v) && v >= 0;
    	},
    	width: function (v) {
    		return !isNaN(v) && v >= 0;
    	},
    	height: function (v) {
    		return !isNaN(v) && v >= 0;
    	},
    	/** any value is okay, empty and unknown ones will fall back to a default*/
    	position: function (v) { return true; },
    	mode: function (v) { return true; },
    };
	//                               NON-MRAID PROPERTIES                               //
	//////////////////////////////////////////////////////////////////////////////////////


	/* adding a startsWith() function to use on strings */
	if (typeof String.prototype.startsWith != 'function') {
		String.prototype.startsWith = function (str) {
			return this.slice(0, str.length) == str;
		};
	}


	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////// BODY OF NON-MRAID FUNCTION IMPLEMENTATION //////////////////////////////////////////////////////////////////////////////
	this.non_mraid = {
		setDefaultPosition: function (jsonString) {
			var jsonObj = JSON.parse(jsonString);

			defaultPosition.x = jsonObj.x;
			defaultPosition.y = jsonObj.y;
			defaultPosition.width = jsonObj.width;
			defaultPosition.height = jsonObj.height;
		},

		setDurationTime: function (milliseconds) {
			this.displayDurationTime = milliseconds;
		},

		startDurationTimer: function () {
			if (this.displayDurationTime > 0) {
				this.displayDurationTimer = window.setTimeout(mraid.close, this.displayDurationTime);
			}
		},

		stopDurationTimer: function () {
			window.clearTimeout(this.displayDurationTimer);
		},

		getDensity: function () {
			return density;
		},

		setDensity: function (dens) {
			if (dens <= 0) {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "Density can't be set to a value <= 0",
					action: "setDensity"
				});
				return;
			}
			density = dens;
		},

		setAbsSize: function (width, height) {
			absSize.width = width;
			absSize.height = height;
		},

		// TODO:-- VJ Legacy implementation. Do we need this function?
		testEvent: function () {
			mraid_bridge.logMessage("fireEvent : ");
		},

		fireEvent: function (event, eventData) {
            //console.log("FIRE EVENT: " + event + " | eventData: " + eventData);
			var listener, _i, _len, _ref, _results;
			if (listeners[event]) {
				_ref = listeners[event];
				_results = [];

				for (_i = 0, _len = _ref.length; _i < _len; _i++) {
					listener = _ref[_i];

					switch (event) {
						case EVENT_SIZECHANGE:
							listener(eventData.width, eventData.height);
							break;
						case EVENT_READY:
							listener();
							break;
						case EVENT_STATECHANGE:
							listener(state);
							break;
						case EVENT_ERROR:
							var message = eventData.message;
							var action = eventData.action;
							listener(message, action);
							break;
						case EVENT_SUCCESS:
							listener(eventData.action);
							break;
						case EVENT_WARNING:
							var message = eventData.message;
							var action = eventData.action;
							listener(message, action);
							break;
						case EVENT_VIEWABLECHANGE:
							_results.push(listener(viewable));
							break;
						case EVENT_EXPOSURECHANGE:
							listener(eventData.exposedPercentage, eventData.visibleRectangle, eventData.occlusionRectangles);
							break;
                        case EVENT_AUDIOVOLUMECHANGE:
                            listener(eventData.volumePercentage);
                            break;
						case EVENT_VISXONSCROLL:
							listener(eventData.initialScrollViewHeight, eventData.currentScrollViewHeight, eventData.currentScrollPosition, eventData.scrollPositionDelta);
							break;
						case EVENT_VISXONBACKPRESS:
							listener();
							break;
						case EVENT_VISX_STICKY_IS_CLOSED:
							listener();
							break;
						case EVENT_VISX_STICKY_IS_DISPLAYED:
							listener();
							break;
						case EVENT_VISX_TO_STICKY:
							listener();
							break;
						case EVENT_VISX_TO_STICKY_CLOSE:
							listener();
							break;
						case EVENT_VISX_TO_STICKY_CLOSE_BUTTON:
							listener();
							break;
						default:
							_results.push(void 0);
							break;
					}
				}
				return _results;
			}
		},

		setState: function (state_in) {
		console.log("SET STATE: " + state_in);
			if (states.indexOf(state_in) < 0) {
				console.log('mraid.js: setState: "' + state_in + '" is not a valid MRAID state, aborting.');
				return;
			}
			state = state_in;
			return non_mraid.fireEvent(EVENT_STATECHANGE);
		},

		setViewable: function (is_viewable) {
			viewable = is_viewable;
			return non_mraid.fireEvent(EVENT_VIEWABLECHANGE);
		},

		exitApplicationAlert: function (bool, alertData) {
			mraid_bridge.setExitApplicationAlertData(bool, alertData.title,
				alertData.text, alertData.cancelLabel, alertData.okLabel);
		},

		setCloseButtonPosition: function (str) {
			mraid_bridge.setCloseButtonPosition(str);
		},

		getAbsoluteScreenSize: function () {
			return absSize;
		},

		validate: function (obj, validators, action, merge) {
			var i, prop;
			if (!merge) {
				// Check to see if any required properties are missing.
				if (obj === null) {
					return false;
				}

				for (i in validators) {
					if (validators.hasOwnProperty(i)
						&& obj[i] === undefined) {
						return false;
					}
				}
			}

			for (prop in obj) {
				var validator = validators[prop];
				var value = obj[prop];
				if (validator && !validator(value)) {
					// Failed validation.
					return false;
				}
			}
			return true;
		},

		hideCustomControls: function () {
			/** TODO:-- VJ Check implementation legacy code and comment is for Android 2.3.x and
				 now the Android 5.0 is minimum
			*/

			/**
			 * Due to a bug in Android 2.3.x transparent areas of a video view
			 * will not be refreshed. Therefore, we have to invalidate the whole
			 * body.
			 */

			var mute = document.getElementById("__mute__"), controls = document
				.getElementById("__controls__"), body = document
					.querySelector('body'), body_opac_org = body.style.opacity;
			if (mute) {
				mute.style.display = "none";
			}
			if (controls) {
				controls.style.display = "none";
			}

			body.style.opacity = 0.99;
			setTimeout(function () {
				body.style.opacity = body_opac_org;
			}, 10);
		},

		isAppInstalled: function (packageName) {
			return add_bridge.isAppInstalled(packageName);
		},

		/**
		* Creative provide html-content, that will be rendered in an additional webView below the
		* creative-webView.
		*
		* @param - html
		*/
		visxShowAdvertisementMessageBelow : function(html) {
            mraid_bridge.visxShowAdvertisementMessageBelow(html);
		},

		/**
		* Creative provide html-content, that will be rendered in an additional webView above the
		* creative-webView.
		*
		* @param - string
		*/
		visxShowAdvertisementMessageAbove : function(string) {
            mraid_bridge.visxShowAdvertisementMessageAbove(string);
		},

		visxEnableOnScrollEvent : function() {
            mraid_bridge.visxEnableOnScrollEvent();
        },

        initAudioVolumeChange : function() {
            mraid_bridge.initAudioVolumeChange();
        },

		initPlacementEffect: function (effectString) {
			placementEffect = effectString;
		},

		initPlacementDimensions: function (webViewWidth, webViewHeight, viewportWidth, viewportHeight) {
			webViewDimension.width = webViewWidth;
			webViewDimension.height = webViewHeight;
			viewportDimension.width = viewportWidth;
			viewportDimension.height = viewportHeight;
		},

		visxSetPlacementDimension: function (webViewWidth, webViewHeight, viewportWidth, viewportHeight) {
		    webViewDimension.width = webViewWidth;
            webViewDimension.height = webViewHeight;
            viewportDimension.width = viewportWidth;
            viewportDimension.height = viewportHeight;
            mraid_bridge.visxSetPlacementDimension(webViewWidth, webViewHeight, viewportWidth, viewportHeight);
		},

		visxGetPlacementDimension: function () {
			return {
				webViewDimension: {
					width: webViewDimension.width,
					height: webViewDimension.height
				},
				viewportDimension: {
					width: viewportDimension.width,
					height: viewportDimension.height
				}
			}
		},

		visxSetPlacementEffect: function (effect, webViewWidth, webViewHeight, viewportWidth, viewportHeight) {
			placementEffect = effect;
			webViewDimension.width = webViewWidth;
			webViewDimension.height = webViewHeight;
			viewportDimension.width = viewportWidth;
			viewportDimension.height = viewportHeight;
			mraid_bridge.visxSetPlacementEffect(effect, webViewWidth, webViewHeight, viewportWidth, viewportHeight);
		},

		visxGetPlacementEffect: function () {
		    return placementEffect;
		},

		visxClosePlacement: function () {
		    mraid_bridge.visxClosePlacement();
		},

		visxClearPlacement: function () {
			mraid_bridge.visxClearPlacement();
		},

		visxRefreshPlacement: function () {
			mraid_bridge.visxRefreshPlacement();
        },

        visxVideoWasUnmuted : function() {
            mraid_bridge.visxVideoWasUnmuted();
        },

        visxVideoWasMuted : function() {
            mraid_bridge.visxVideoWasMuted();
        },

        visxVideoWasCanceled : function() {
            mraid_bridge.visxVideoWasCanceled();
        },

        visxVideoFinished : function() {
            mraid_bridge.visxVideoFinished();
		},

		setLandingPageURLs: function (url) {
            if (!url || url == "") {
            	non_mraid.fireEvent(EVENT_ERROR, {
            		message: "The URL is undefined",
            		action: "setLandingPageURLs()"
            	});
            	return;
            }
            mraid_bridge.setLandingPageURLs(url);
        },

        visxShowBrandedTakeoverSticky : function(html, width, height, animation, direction) {
            mraid_bridge.visxShowBrandedTakeoverSticky(html, width, height, animation, direction);
		},

        visxHideBrandedTakeoverSticky : function(animation, direction) {
            mraid_bridge.visxHideBrandedTakeoverSticky(animation, direction);
		},

		setCurrentAppOrientation : function (orientation, locked) {
            currentAppOrientationProperties.orientation = orientation;
            currentAppOrientationProperties.locked = locked == 1 ? true : false;
        },

        getCurrentAppOrientation : function () {
            return currentAppOrientationProperties;
        },

		setCanUseOnScrollEvent: function (canUseScrollEvent) {
			canUseOnScrollEvent = canUseScrollEvent;
		},

		getCanUseOnScrollEvent: function () {
			return canUseOnScrollEvent;
		},

		setDefaultStickyProperties: function (thresholdSDK, widthSDK, heightSDK, positionSDK, modeSDK) {
            stickyProperties = {
                threshold: thresholdSDK,
                width: widthSDK,
                height: heightSDK,
                position: positionSDK,
                mode: modeSDK
            }
		},

		visxToSticky : function() {
            mraid_bridge.visxToSticky();
        },

		visxCloseSticky : function() {
            mraid_bridge.visxCloseSticky();
        },

		getVisxApiVersion: function () {
			return apiVersion;
		},
	}

	//////////////////////////////////////////////////////////////////////////////// BODY OF NON-MRAID FUNCTION IMPLEMENTATION //////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////// MAIN BODY OF THE MRAID CREATIVE ////////////////////////////////////////////////////////////////////////////////////
	this.mraid = {

		//////////////////////////////////////////////////////////////////////////////////////
		//                                 MRAID 3.0 METHODS                                //

		/**
		* The version number indicates the version of MRAID that the host
		* supports (1.0, 2.0, or 3.0, etc.), NOT the version of the vendor's SDK.
		*
		* MRAID v3.0, p. 34
		*/
		getVersion: function () {
			return mraidVersion;
		},

		/**
		* The ad calls addEventListener() to register a specific listener for a specific
		* event. The ad may register more than one listener, each to support listening for
		* separate event. The host dispatches an event to all registered listeners for each
		* specific event that occurs. The ad may also register a single listener to multiple
		* events instead of a listener for each event
		*
		* MRAID v3.0, p. 34
		*/
		addEventListener: function (event, listener) {
			if (!event || !listener) {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "Event or/and Listener are undefined",
					action: "addEventListener"
				});
				return;
			}

			if (event === EVENT_READY || event === EVENT_STATECHANGE
				|| event === EVENT_VIEWABLECHANGE
				|| event === EVENT_ERROR
				|| event === EVENT_SUCCESS
				|| event === EVENT_SIZECHANGE
                || event === EVENT_EXPOSURECHANGE
                || event === EVENT_AUDIOVOLUMECHANGE
                || event === EVENT_VISXONSCROLL
                || event === EVENT_VISXONBACKPRESS
                || event === EVENT_VISX_STICKY_IS_CLOSED
                || event === EVENT_VISX_STICKY_IS_DISPLAYED
                || event === EVENT_VISX_TO_STICKY
                || event === EVENT_VISX_TO_STICKY_CLOSE
                || event === EVENT_VISX_TO_STICKY_CLOSE_BUTTON) {

				if (!listeners[event]) {
					listeners[event] = [];
				}
				var index = listeners[event].indexOf(listener);
				/* Only add the listener if it is not present in the array yet */
				if (index == -1) {
					listeners[event].push(listener);
				}
				/* When audioVolumeChange is added, poll manually */
				if (event === EVENT_AUDIOVOLUMECHANGE) {
				    non_mraid.initAudioVolumeChange();
				}
			} else {
				non_mraid.fireEvent(EVENT_ERROR,
					{
						message: "Unknown event: " + event,
						action: "addEventListener"
					});
			}
		},

		/**
		* When the ad no longer needs notification of a particular event,
		* removeEventListener() is used to unregister to that event. To avoid errors,
		* event listeners must always be removed when they are no longer needed. If no
		* listener function is specified in the in the listener attribute for the call, then all
		* functions listening to the event will be removed.
		*
		* MRAID v3.0, p. 35
		*/
		removeEventListener: function (event, listener) {

			/* Event or/and Listener are undefined */
			if (!event || !listeners[event]) {
				/* Request dropped if event not provided or listener not registered */
				/*non_mraid.fireEvent(EVENT_ERROR, {
					message: "Event or/and Listener are undefined",
					action: "removeEventListener"
				});*/

			} else {
				/* All listeners for the given event should be removed */
				if (!listener) {
					listeners[event] = null;
					delete listeners[event];

					//TODO:-- VJ Commented legacy code. Do we need it?
					/*non_mraid.fireEvent(EVENT_ERROR, {
					message: "Event or/and Listener are undefined",
					action: "removeEventListener"
					});*/

					/* Only the provided listener should be removed */
				} else {
					/* Check if the listener is already registered. */
					var index = listeners[event].indexOf(listener);

					/* If listener already stored, remove it. */
					if (index != -1) {
						listeners[event].splice(index, 1);
						if (listeners[event].length == 0) {
							listeners[event] = null;
							delete listeners[event];
						}
					}
				}
			}
		},

		/**
		* The ad can call the open() method to prompt the host to open an external
		* mobile website in a browser window that is the default browser on the user’s
		* device The purpose of this method is to handle clickthroughs in the ad. All
		* MRAID ads must handle clickthroughs using the open() method.
		*
		* MRAID v3.0, p. 35
		*/
		open: function (url) {
		console.log("MRAID.open() url: " + url);
			if (!url || url == "") {
				console.log("MRAID.open() EVENT_ERROR");
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "The URL is undefined",
					action: "open"
				});
				return;
			}
			return mraid_bridge.open(url);
		},

		/**
		* The ad uses close() to downgrade the container state. The host responds by
		* changing the state depending on the current state using the stateChange event
		*
		* For ads in an expanded or resized state, the close() method moves the ad to a
		* default state. For interstitial ads in a default state, the close() method moves to
		* a hidden state.
		*
		* For ads in a default state, ad developers must avoid using mraid.close() . To
		* inform the container that the ad needs to be dismissed, use mraid.unload().
		* If an ad uses multiple resize() calls or a resize() followed by the expand()
		* call, close() returns the container state to 'default.' Calling close() in these
		* instances does NOT simply undo the most recently called resize() or
		* expand().
		*
		* MRAID v3.0, p. 36
		*/
		close: function () {
			return mraid_bridge.close();
		},

		/**
		* *** DEPRECATED ***
		*/
		useCustomClose: function (bool) {
			mraid_bridge.useCustomClose();
		},

		/**
		* The ad uses expand() to request support for an ad expansion experience, either
		* by changing the width and height of the current webview (one-part creative) or by
		* opening a new webview at the highest level in an expanded size (two-part
		* creative). The expanded view can either contain a new HTML document if a URL
		* is specified, or it can reuse the same document that was in the default position.
		*
		* MRAID v3.0, p. 39
		*/
		expand: function () {
			return mraid_bridge.expand();
		},

		/*
		* *** DEPRECATED ***
		*
		* The isViewable() method is deprecated in MRAID 3.0 and will be removed in
		* future versions of MRAID. This feature remains in MRAID 3.0 for the purposes of
		* backward compatibility.
		* The isViewable() method returns whether the ad container is currently on or off
		* the screen. The viewableChange event fires when the ad moves from on-screen
		* to off-screen and vice versa.
		*
		* MRAID v3.0, p. 39
		*/
		isViewable: function () {
			return viewable;
		},

		/**
		* The ad uses playVideo() when a video component of the creative needs to
		* play in the device's native player. To play the video inline rather than in the native
		* player, the ad designer must use HTML5 video tags instead.
		* MRAID v3.0, p. 41
		*/
		playVideo: function (URI) {
			if (!URI || typeof URI !== 'string' || URI.trim().length <= 0) {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "Video URI is empty",
					action: "playVideo"
				});
				return;
			}
			return mraid_bridge.playVideo(URI);
		},

		/**
		* Calling resize() is a request for a container size change that accommodates
		* the creative size change. Resize is used for a succession of changes or a size
		* change that is less than full screen size and that doesn't interfere with app
		* operation.
		*
		* MRAID v3.0, p. 41
		*/
		resize: function () {
			if (resizeProperties === null) {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: 'Resize properties have not been set.',
					action: 'resize'
				});
				return;
			}

			if (state == STATE_EXPANDED) {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "Resize aborted, current state is 'expanded'.",
					action: 'resize'
				});
				return;
			}
			return mraid_bridge.resize();
		},

		/**
		* Stores a picture to the device after requesting user permission.
		* An ad that offers users the option to store an image, storePicture() can be
		* used (if supported), to place an image in the device's photo album. The image
		* may be part of the ad or retrieved from a server.
		* @param url		the URL of the picture to store
		* @param msg		the message displayed to the user asking for permission; optional, non-MRAID
		* @param lblAllow	the label of the "allow" button; optional, non-MRAID
		* @param lblDeny	the label of the "deny" button; optional, non-MRAID
		*
		* MRAID v3.0, p. 42
		*/
		// TODO:-- VJ remove unused non-MRAID parameters
		storePicture: function (url, msg, lblAllow, lblDeny) {

			if (!this.supports('storePicture')) {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "MRAID storePicture is not supported.",
					action: "storePicture"
				});
				return;
			}

			if (url === undefined || url === null || typeof url !== 'string') {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "Image URL must be of type 'string'",
					action: "storePicture"
				});
				return;
			}

			if (msg && msg !== '' && lblAllow && lblAllow !== '' && lblDeny && lblDeny !== '') {
				mraid_bridge.storePicture(url, msg, lblAllow, lblDeny);
				return;
			}
			mraid_bridge.storePicture(url);
		},

		/**
		 * Opens the device UI to create a new calendar event.
		 * Invokes the device's native "create calendar event" sheet pre-populated with data supplied by the ad.
		 * @param parameters	JavaScript Object {...} – this object contains the parameters for the calendar entry,
		 * written according to the W3C specification for calendar entries.
		 *
		 * MRAID v3.0, p. 42
		 */
		createCalendarEvent: function (parameters) {
			var start = "";

			if (!parameters.start) {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "The start date and time of the event must be provided as a valid date or time string.",
					action: "createCalendarEvent"
				});
				return;
			} else {
				start = parameters.start;
			}

			var description = "";

			if (!parameters.description) {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "The description of the event must be provided as a valid string.",
					action: "createCalendarEvent"
				});
				return;
			} else {
				description = parameters.description;
			}

			var id = 		    (parameters.id) 		    ? parameters.id 		    : "";
			var location = 		(parameters.location) 		? parameters.location 		: "";
			var summary = 		(parameters.summary) 		? parameters.summary 		: "";
			var end = 			(parameters.end) 			? parameters.end 			: "";
			var status =		(parameters.status) 		? parameters.status 		: "";
			var transparency =	(parameters.transparency) 	? parameters.transparency 	: "";
			var recurrence = 	(parameters.recurrence) 	? JSON.stringify(parameters.recurrence) 	: "";
			var reminder = 		(parameters.reminder) 		? parameters.reminder 		: "";

			mraid_bridge.createCalendarEvent(id, description, location, summary,
			 start, end, status, transparency, recurrence, reminder);
		},

		/*
		* The key purpose of this method is to allow the ad to communicate to the host that
		* the ad must no longer be shown to the user. The ad can request the host to
		* completely dismiss the ad by using the unload() method. The host responds by
		* dismissing the ad and then either removing the webview or replacing it with
		* another document or refreshing it with a new ad.
		* MRAID v3.0, p. 38
		*/
		unload: function () {
	        mraid_bridge.unload();
		},

		// TODO MRAID 3.0 METHODS to be implemented ------------------------------------------

		/**
		* Once the ad has checked for VPAID support in the mraid.supports() method,
		* it needs to give the container a vpaidObject that the container uses to
		* communicate events and information about the video playback. The VPAID
		* object enables the host to subscribe to VPAID events, start the ad, and call a
		* limited set of VPAID methods. After initVpaid() is called, in order for video
		* playback to begin, the container must call vpaidObject.startAd(). The ad
		* must not begin video playback until videoObject.startAd() is called.
		*
		* MRAID v3.0, p. 44
		*/
		initVpaid: function (vpaidObjectParam) {
			vpaidObject = vpaidObjectParam;
			// TODO Implerment with MRAID 3.0 Specs
			// return mraid_bridge.initVpaid(vpaidObject);
		},

		//                                MRAID 3.0 METHODS                                 //
		//////////////////////////////////////////////////////////////////////////////////////

		//////////////////////////////////////////////////////////////////////////////////////
		//                              MRAID 3.0 PROPERTIES                                //

		/**
		* Returns if a MRAID 3.0 feature is supported by the device.
		*
		* MRAID v3.0, p. 46-47
		*/
		supports: function (feature) {
			return supportedFeatures[feature] === true;
		},

		/**
		* The ad calls getPlacementType() to determine whether it's being loaded in an
		* inline placement or an interstitial.
		*
		* MRAID v3.0, p. 47
		*/
		getPlacementType: function () {
			return placementType;
		},

		/**
		* The ad uses the orientationProperties to query the host about the current
		* orientation of the device and set orientationProperties to accommodate ad
		* display for expansions and interstitials. A banner in its default state cannot use
		* orientationProperties
		*
		* MRAID v3.0, p. 47
		*/
		setOrientationProperties: function (properties) {
			/* If neither allowOrientationChange nor forceOrientation is set, throw an error (properties are invalid) */
			if (properties.allowOrientationChange === undefined && !properties.forceOrientation) {
				non_mraid.fireEvent(EVENT_ERROR, {
					message: "Invalid set of properties",
					action: "setOrientationProperties"
				});
				return;
			}

			if (properties.allowOrientationChange !== undefined) {
				orientationProperties.allowOrientationChange = properties.allowOrientationChange;
			}
			/* If forceOrientation is set, but not to one of the predefined values, throw an error */
			if (properties.forceOrientation) {
				if (properties.forceOrientation != FORCEORIENTATION_PORTRAIT &&
					properties.forceOrientation != FORCEORIENTATION_LANDSCAPE &&
					properties.forceOrientation != FORCEORIENTATION_NONE) {
					non_mraid.fireEvent(EVENT_ERROR, {
						message: "Invalid argument for forceOrientation: " + properties.forceOrientation,
						action: "setOrientationProperties"
					});
					return;
				}
				orientationProperties.forceOrientation = properties.forceOrientation;
			}

			/* Actually setting orientation properties by calling native Java-method */
			mraid_bridge.setOrientationProperties(orientationProperties.allowOrientationChange, orientationProperties.forceOrientation);
		},

		/**
		* The getOrientationProperties method prompts the host to return the whole
		* orientationProperties object.
		*
		* MRAID v3.0, p. 49
		*/
		getOrientationProperties: function () {
			return orientationProperties;
		},

		/**
		* The ad calls getCurrentPosition to query the host for the current position of
		* the ad container.
		* The host returns the current position and size of the ad container, measured in
		* density-independent pixels.
		*
		* MRAID v3.0, p. 50
		*/
		getCurrentPosition: function () {
			var jsonString = mraid_bridge.getCurrentPosition();
			var jsonObj = JSON.parse(jsonString);
			return {
				x: jsonObj.x,
				y: jsonObj.y,
				width: jsonObj.width,
				height: jsonObj.height
			};
		},

		/**
		* The ad calls getDefaultPosition query the host about the position and size of
		* the default ad container, measured in density-independent pixels.
		* The host returns values for the default position and size of the ad container
		* regardless of what state the calling view is in.
		*
		* MRAID v3.0, p. 50
		*/
		getDefaultPosition: function () {
			return defaultPosition;
		},

		/**
		* At any time, the ad may use getState to query the host about the state of the ad
		* container and make requests accordingly.
		* The host returns the current state of the ad container using values that describe
		* whether the ad container is in its default and fixed position, in an expanded or
		* resized state, a larger position, or hidden.
		*
		* MRAID v3.0, p. 51
		*/
		getState: function () {
			return state;
		},

		/**
		* The ad uses getExpandProperties to query the host on the current expand
		* settings, and setExpandProperties to set the width and height for an
		* expansion along with optionally specifying the use of a custom close indicator
		*
		* MRAID v3.0, p. 51
		*/
		setExpandProperties: function (properties) {
			var desiredProperties = ['width', 'height', 'useCustomClose'];
			var i;
			var length = desiredProperties.length;
			for (i = 0; i < length; i++) {
				var propname = desiredProperties[i];
				if (properties.hasOwnProperty(propname)) {
					expandProperties[propname] = properties[propname];
				}
			}
			return mraid_bridge.setExpandProperties(expandProperties['width'], expandProperties['height'], expandProperties['useCustomClose']);
		},

		/**
		* See setExpandProperties
		*
		* MRAID v3.0, p. 51
		*/
		getExpandProperties: function () {
			return expandProperties;
		},

		/**
		* The ad calls getMaxSize to query the host about the maximum size (in densityindependent
		* pixels) to which the ad may resize
		*
		* MRAID v3.0, p. 53
		*/
		getMaxSize: function () {
			return JSON.parse(mraid_bridge.getMaxSize());
		},

		/**
		* The ad calls getScreenSize to query the host about the dimensions of the
		* device screen size, especially before expanding an ad (for resizing use
		* getMaxSize). The host returns current device screen width and height, in
		* density-independent pixels.
		*
		* MRAID v3.0, p. 53
		*/
		getScreenSize: function () {
			return JSON.parse(mraid_bridge.getScreenSize());
		},

		/**
		* Before the ad executes a resize, it can call getResizeProperties to query the
		* host about current settings for resizing the ad container. Then the ad can call
		* setResizeProperties to update the resizeProperties object. The host uses
		* the properties set in this object to adjust the ad container when the ad calls
		*
		* MRAID v3.0, p. 54
		*/
		setResizeProperties: function (properties) {
			var i;
			if (mraid.validate(properties, resizePropertyValidators,
				'setResizeProperties', true)) {
				var desiredProperties = ['width', 'height', 'offsetX',
					'offsetY', 'customClosePosition', 'allowOffscreen'];
				var tmpResizeProps = {};
				var length = desiredProperties.length;
				for (i = 0; i < length; i++) {
					var propname = desiredProperties[i];
					if (properties.hasOwnProperty(propname)) {
						tmpResizeProps[propname] = properties[propname];
					}
				}
			    tmpResizeProps['customClosePosition'] = 'top-right';
				resizeProperties = tmpResizeProps;
				mraid_bridge.setResizeProperties(resizeProperties['width'],
					resizeProperties['height'],
					resizeProperties['customClosePosition'],
					resizeProperties['offsetX'],
					resizeProperties['offsetY'],
					resizeProperties['allowOffscreen']);
			} else {
				mraid.fireEvent(EVENT_ERROR,
					{
						message: "The given set of resize properties is not valid.",
						action: "setResizeProperties"
					}
				);
			}
		},

		/**
		* See setResizeProperties
		*
		* MRAID v3.0, p. 54
		*/
		getResizeProperties: function () {

			/** "The getResizeProperties method returns the whole JavaScript object resizeProperties object.",
			 * and if there isn't any, we return null. */
			if (!resizeProperties) return null;

			var properties = {
				width: resizeProperties.width,
				height: resizeProperties.height,
				offsetX: resizeProperties.offsetX,
				offsetY: resizeProperties.offsetY,
				customClosePosition: resizeProperties.customClosePosition,
				allowOffscreen: resizeProperties.allowOffscreen
			};
			return properties;
		},

		/**
		* Knowing the location of the device when an ad is active can enhance the creative
		* possibilities for an ad, but not all apps support access to location details. The
		* supports feature in MRAID 3.0 (see section 6.1) identifies whether "location" is
		* supported.
		*
		* MRAID v3.0, p. 49
		*/
		getLocation: function () {
	        return JSON.parse(mraid_bridge.getLocation());
		},

		/**
		* The ad calls getCurrentAppOrientation to query the host about the current
		* orientation of the app.
		* The host returns the current orientation of the app as either "portrait" or
		* "landscape" and whether or not the given orientation is locked. If locked, then the
		* forceOrientation option in setOrientationProperties is not supported.
		*
		* MRAID v3.0, p. 49
		*/
		getCurrentAppOrientation: function () {
	        return JSON.parse(mraid_bridge.getCurrentAppOrientation());
		},

		//                              MRAID 3.0 PROPERTIES                                //
		//////////////////////////////////////////////////////////////////////////////////////

		//////////////////////////////////////////////////////////////////////////////////////
		//                                MRAID 3.0 EVENTS                                  //

		/**
		* The exposure change event is not supported in two-part ads.
		* When the ad registers a listener for the exposureChange event, the host sends
		* the event asynchronously with the initial exposure state to all listeners for the ad.
		* After the initial event, the host reports exposureChange events anytime the
		* exposed area changes. An exposedPercentage value of 0.0 (zero) indicates the
		* current ad container is not in view or has been moved to the background.
		* Exposure changes can occur when the ad container or any of its parent views is
		* scrolled, translated, or clipped. The host also reports exposureChange events
		* when the visibility of the ad changes because of application or UI transitions.
		*
		* MRAID v3.0, p. 60
		*/
		exposureChange: function (exposedPercentage, visibleRectangle, occlusionRectangles) {
			mraid.fireEvent('exposureChange', { 'exposedPercentage': exposedPercentage, 'visibleRectangle': visibleRectangle, 'occlusionRectangles': occlusionRectangles });
		},

		/**
		* The exposure change event is not supported in two-part ads.
		* When the ad registers a listener for the exposureChange event, the host sends
		* the event asynchronously with the initial exposure state to all listeners for the ad.
		* After the initial event, the host reports exposureChange events anytime the
		* exposed area changes. An exposedPercentage value of 0.0 (zero) indicates the
		* current ad container is not in view or has been moved to the background.
		* Exposure changes can occur when the ad container or any of its parent views is
		* scrolled, translated, or clipped. The host also reports exposureChange events
		* when the visibility of the ad changes because of application or UI transitions.
		*
		* MRAID v3.0, p. 60
		*/
		audioVolumeChange: function (volumePercentage) {
		    mraid.fireEvent('audioVolumeChange', {'volumePercentage': volumePercentage});
		},

		//                                MRAID 3.0 EVENTS                                  //
		//////////////////////////////////////////////////////////////////////////////////////

		//////////////////////////////////////////////////////////////////////////////////////
		//                              NON-MRAID 3.0 EVENTS                                //

		visxOnScroll: function (initialScrollViewHeight, currentScrollViewHeight, currentScrollPosition, scrollPositionDelta) {
        	mraid.fireEvent('visxOnScroll', { 'initialScrollViewHeight': initialScrollViewHeight, 'currentScrollViewHeight': currentScrollViewHeight, 'currentScrollPosition': currentScrollPosition,  'scrollPositionDelta': scrollPositionDelta});
        },

		//                              NON-MRAID 3.0 EVENTS                                //
		//////////////////////////////////////////////////////////////////////////////////////

		//////////////////////////////////////////////////////////////////////////////////////
		//                                NON-MRAID METHODS                                 //
		/* deprecated, this.mraid should only contain methods belonging to
		* the mraid interface definition  */


		initPlacementEffect: function (effectString) {
			return non_mraid.initPlacementEffect(effectString);
		},

		initPlacementDimensions: function (webViewWidth, webViewHeight, viewportWidth, viewportHeight) {
			return non_mraid.initPlacementDimensions(webViewWidth, webViewHeight, viewportWidth, viewportHeight);
		},

		visxSetPlacementDimension: function (webViewWidth, webViewHeight, viewportWidth, viewportHeight) {
			return non_mraid.visxSetPlacementDimension(webViewWidth, webViewHeight, viewportWidth, viewportHeight);
		},

		visxGetPlacementDimension: function () {
			return non_mraid.visxGetPlacementDimension();
		},

		visxSetPlacementEffect: function (effect, webViewWidth, webViewHeight, viewportWidth, viewportHeight) {
			return non_mraid.visxSetPlacementEffect(effect, webViewWidth, webViewHeight, viewportWidth, viewportHeight);
		},

		visxGetPlacementEffect: function () {
			return non_mraid.visxGetPlacementEffect();
		},

		visxClosePlacement: function () {
			return non_mraid.visxClosePlacement();
		},

		visxClearPlacement: function () {
			return non_mraid.visxClearPlacement();
		},

        visxRefreshPlacement: function () {
            if(state == STATE_DEFAULT || state == STATE_HIDDEN){
                return non_mraid.visxRefreshPlacement();
            } else {
                mraid.fireEvent(EVENT_ERROR,
                    {
                        message: "visxRefreshPlacement requires state to be default or hidden.",
                        action: "visxRefreshPlacement"
                    }
                );
            }
        },

        visxVideoWasUnmuted : function() {
            return non_mraid.visxVideoWasUnmuted();
        },

        visxVideoWasMuted : function() {
            return non_mraid.visxVideoWasMuted();
        },

        visxVideoWasCanceled : function() {
            return non_mraid.visxVideoWasCanceled();
        },

        visxVideoFinished : function() {
            return non_mraid.visxVideoFinished();
		},

		visxShowAdvertisementMessageBelow : function(html) {
            return non_mraid.visxShowAdvertisementMessageBelow(html);
        },

		visxShowAdvertisementMessageAbove : function(string) {
            return non_mraid.visxShowAdvertisementMessageAbove(string);
        },

        visxEnableOnScrollEvent: function () {
        	return non_mraid.visxEnableOnScrollEvent();
        },

		setDensity: function (dens) {
			return non_mraid.setDensity(dens);
		},

		setVolumeValue: function (vol) {
			return non_mraid.setVolumeValue(vol);
		},

		getDensity: function () {
			return non_mraid.getDensity();
		},

		testEvent: function () {
			return non_mraid.testEvent();
		},

		fireEvent: function (event, eventData) {
			return non_mraid.fireEvent(event, eventData);
		},

		setState: function (state_id) {
			return non_mraid.setState(state_id);
		},

		setViewable: function (is_viewable) {
			return non_mraid.setViewable(is_viewable);
		},

		openModalVCWithURL: function (URL) {
			return non_mraid.openModalVCWithURL(URL);
		},

		exitApplicationAlert: function (bool, alertData) {
			return non_mraid.exitApplicationAlert(bool, alertData);
		},

		setCloseButtonPosition: function (str) {
			return non_mraid.setCloseButtonPosition(str);
		},

		getAbsoluteScreenSize: function () {
			return non_mraid.getAbsoluteScreenSize();
		},

		validate: function (obj, validators, action, merge) {
			return non_mraid.validate(obj, validators, action, merge);
		},

		setDimmingViewProperties: function (color, opacity) {
		},

		hideCustomControls: function () {
			return non_mraid.hideCustomControls();
		},

		getExposureChangeAvailability: function () {
			return isExposureChangeEnabled;
		},

		setIsMediationAdView: function (isMediationFlag) {
			isMediationAdView = isMediationFlag;
		},

		getIsMediationAdView: function () {
			return isMediationAdView;
		},

		setLandingPageURLs: function (urls) {
			return non_mraid.setLandingPageURLs(urls);
		},

		visxShowBrandedTakeoverSticky : function(html, width, height, animation, direction) {
            return non_mraid.visxShowBrandedTakeoverSticky(html, width, height, animation, direction);
        },

		visxHideBrandedTakeoverSticky : function(animation, direction) {
            return non_mraid.visxHideBrandedTakeoverSticky(animation, direction);
        },

        setIsUniversalAd: function (val) {
        	isUniversalAd = val;
        },

		getIsUniversalAd: function () {
			return isUniversalAd;
		},

        setCurrentAppOrientation: function (orientation, locked) {
        	return non_mraid.setCurrentAppOrientation(orientation, locked)
        },

		getCurrentAppOrientation: function () {
			return non_mraid.getCurrentAppOrientation();
		},

		setCanUseOnScrollEvent: function (canUseScrollEvent) {
		    return non_mraid.setCanUseOnScrollEvent(canUseScrollEvent)
		},

		canUseOnScrollEvent: function () {
			return non_mraid.getCanUseOnScrollEvent();
		},

		visxGetStickyProperties: function () {
			/** "The getStickyProperties method returns the whole JavaScript object stickyProperties object.",
			 * and if there isn't any, we return null. */
			if (!stickyProperties) return null;

			var properties = {
				threshold: stickyProperties.width,
				width: stickyProperties.width,
				height: stickyProperties.height,
				position: stickyProperties.position,
				mode: stickyProperties.mode,
			};
			return properties;
		},

		setDefaultStickyProperties: function (thresholdSDK, widthSDK, heightSDK, positionSDK, modeSDK) {
			return non_mraid.setDefaultStickyProperties(thresholdSDK, widthSDK, heightSDK, positionSDK, modeSDK);
		},

		visxSetStickyProperties: function (properties) {
			var i;
			if (mraid.validate(properties, stickyPropertyValidators,
				'setStickyProperties', true)) {
				var desiredProperties = ['threshold', 'width', 'height', 'position', 'mode'];
				var tmpStickyProps = {};
				var length = desiredProperties.length;
				for (i = 0; i < length; i++) {
					var propname = desiredProperties[i];
					if (properties.hasOwnProperty(propname)) {
						tmpStickyProps[propname] = properties[propname];
					}
				}
				stickyProperties = tmpStickyProps;
				mraid_bridge.visxSetStickyProperties(
				    stickyProperties['threshold'],
				    stickyProperties['width'],
					stickyProperties['height'],
					stickyProperties['position'],
					stickyProperties['mode']);
			} else {
				mraid.fireEvent(EVENT_ERROR,
					{
						message: "The given set of sticky properties is not valid.",
						action: "setStickyProperties"
					}
				);
			}
		},

		visxToSticky : function() {
            return non_mraid.visxToSticky();
        },

		visxCloseSticky : function() {
            return non_mraid.visxCloseSticky();
        },

		getVisxApiVersion: function () {
			return non_mraid.getVisxApiVersion();
		},
		//                                NON-MRAID METHODS                                 //
		//////////////////////////////////////////////////////////////////////////////////////
	};
	//////////////////////////////////////////////////////////////////////////////////// MAIN BODY OF THE MRAID CREATIVE ////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////** an object providing methods that are only available internally (e. g. from the native context) *////////////////////////////////////////////////
	this.internal = {

		/** Sets support flags for all native features defined in MRAID v3.0. (p. 46-47).
		 * @param settings 	an object whose keys depict the feature values according to MRAID; values depict the setting
		 */
		setFeatureSupport: function (settings) {
			var keys = Object.keys(settings);

			for (var i = 0; i < keys.length; i++) {
				if (FEATURES.indexOf(keys[i]) < 0) continue;
				supportedFeatures[keys[i]] = settings[keys[i]];
			}
		},

		/** Use this method to initialize the expand properties object. This method
		 * differs from mraid.setExpandProperties in that it does not delegate its invocation
		 * to the native layer. 
		 */
		setExpandProperties: function (width, height, useCustomClose) {
			expandProperties.width = width;
			expandProperties.height = height;
			expandProperties.useCustomClose = !!useCustomClose;
			expandProperties.isModal = true;
		},

		/** Sets the placement type that will be available through mraid.getPlacementType.*/
		setPlacementType: function (type) {
			if (['inline', 'interstitial'].indexOf(type) < 0) return;
			placementType = type;
		}
	};
	//////////////////////////////////////////////////////** an object providing methods that are only available internally (e. g. from the native context) *////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Overridden window.open() function for some edge case creatives not calling mraid.open()
     */
    window.open = function (url) {
        mraid.open(url);
    }
}).call(this);
